/*
 *  CWFtp.C - Main part of cwftp
 *  Copyright (C) 1995 Christian Wolf
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *  You can reach me, the author of cwftp as <chris@MedIS.DE>
 *
 * $Id: CWFtp.C,v 1.4 1996/05/13 17:34:46 chris Exp $
 *
 * $Log: CWFtp.C,v $
 * Revision 1.4  1996/05/13 17:34:46  chris
 * *** empty log message ***
 *
 * Revision 1.3  1996/05/13 16:49:54  chris
 * STOR, Bugfixes
 *
 * Revision 1.2  1996/01/30  09:26:04  chris
 * *** empty log message ***
 *
 * Revision 1.1  1995/05/08  11:47:52  chris
 * Initial revision
 *
 *
 */

#define VERSION "1.0.1"

#include "Ftp.H"
#include "Split.H"
#include "TimeVal.H"
#include "RegExp.H"

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <utime.h>   /* utime */
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
extern char *optarg;
extern int optind;
extern int opterr;
#endif

int verbose = 0; /* be more verbose */
int warnings= 0; /* print warnings */
int saveold = 0; /* save old files before getting a new version */
int newonly = 0; /* get only non existent files (ignore date) */
int deloldf = 0; /* delete old files (no longer on original server) */
int subdirs = 0; /* recurse into subdirectories */
int force   = 0; /* always get a new version (ignore date) */
int dontget = 0; /* do not really get anything, just test */
int nodelete= 0; /* save old files (do not remove nor rename) */
int do_transmit=0; /* transmit files to server instead of retrieving */

tRegularExpression AsciiPattern( NULL );
tRegularExpression IncludePattern( NULL );
tRegularExpression ExcludePattern( NULL );
tRegularExpression KeepPattern( NULL );

/* ============================================================ main */

inline char *mydup( char *ptr )
{
	char *np = new char[ strlen(ptr) + 1 ];
	strcpy( np, ptr );
	return np;
}

void makedirstr( char *newdir, char *basedir, char *subdir )
{
	int len;

	while( *subdir == '/' || *subdir == '.' ) subdir++;

	strcpy( newdir, basedir );
	if( *subdir != '\0' )
	{
		len = strlen( newdir );
		if( newdir[len-1] != '/' ) newdir[len++] = '/';
		strcpy( &newdir[len], subdir );
	}
}


class MirrorSubDir {
	public:
	char *dir;

	MirrorSubDir (char *s) { dir = mydup (s); };
	~MirrorSubDir () { if (dir) free (dir); };
};


class transfer_stats {
	public:
	int total;		// total of dir entries
	int dirs;		// directories
	int f_ignored;	// files ignored
	int f_got;		// files retrieved
	int f_ascii;	// files retrieved as ascii instead of binary (AsciiPattern)
	int f_new;		// new files
	int f_mod;		// modified files
	int f_same;		// files are identical
	int f_rem;		// files not on server (but on local disk), removed/renamed
	int f_kept;		// files kept although not on remote (KeepPattern)

	transfer_stats() { 
		total=dirs=f_ignored=f_got=f_ascii=f_new=f_mod=f_same=f_rem=f_kept=0; };
	void print(char *dir);	// print stats for dir
};

void transfer_stats::print(char *) {
	printf ("  %d/%d files/dirs, %d/%d/%d new/mod/same, %d/%d kept/removed, %d ascii.\n", 
		total-dirs, dirs, 
		f_new, f_mod, f_same, 
		f_kept, f_rem,
		f_ascii);
}

int transmit( tFTP *ftp, char *localdir, char *remotedir, char *dir )
{
	tCWList<tFtpDirNode> LocDirList;
	tFtpDirNode *LocNode;
	char locfile[256];
	char remfile[256];
	char oldfile[256];
	char locdir[256], remdir[256];

	makedirstr( locdir, localdir, dir );
	makedirstr( remdir, remotedir, dir );


	if( ftp->LocalDir( locdir, LocDirList ) == -1 )
	{
		if (warnings >= 0)
			fprintf( stderr, "Error reading local directory `%s'!\n", locdir );
		return -1;
	}

	printf ("%s:\n", dir);

	FOR_CWLIST( LocDirList, LocNode )
	{
		if( LocNode->FDN_Name != NULL
		 && LocNode->FDN_Mode != 0
		 && LocNode->FDN_Name[0] != '.' )
		{
			sprintf( locfile, "%s/%s", locdir, LocNode->FDN_Name );
			sprintf( remfile, "%s/%s", remdir, LocNode->FDN_Name );
			sprintf( oldfile, "%s/.old.%s", locdir, LocNode->FDN_Name );

			printf( "%s -> %s (%s)\n", locfile, remfile, oldfile );

			if( ftp->Put( remfile, locfile, 0 ) )
			{
				fprintf( stderr, "error transmitting '%s'\n", remfile );
			}
			else
			{
				if( deloldf )
				{
					if( remove( locfile ) )
					{
						perror( locfile );
					}
				}
				else
				{
					if( rename( locfile, oldfile ) )
					{
						perror( oldfile );
					}
				}
			}
		}
	}
	return 0;
}

int mirror( tFTP *ftp, char *localdir, char *remotedir, char *dir, int newfiles )
{
	tCWList<tFtpDirNode> RemDirList;
	tCWList<tFtpDirNode> LocDirList;
	tFtpDirNode *RemNode;
	tFtpDirNode *LocNode;
	
	tCWList<MirrorSubDir> SubDirList;
	MirrorSubDir	*subdir;

	char locfile[256];
	char remfile[256];
	char oldfile[256];
	char newfile[256];
	char delfile[256];
	struct stat sbuf;
	struct stat newfilestat;
	long newfilelength;
	struct utimbuf utbuf;
	int getit;
	int reget;
	int ret;
	int sterr;
	int asascii;
	time_t tmpt;

	transfer_stats tfs;
	char locdir[256], remdir[256], s[256];

	makedirstr( locdir, localdir, dir );
	makedirstr( remdir, remotedir, dir );

	if( ftp->RemoteDir( remdir, RemDirList ) == -1 )
	{
		if (warnings >= 0)
			fprintf( stderr, "Error reading remote directory `%s'!\n", remdir );
		return -1;
	}

	if( access( locdir, F_OK ) )
	{
		mkdir( locdir, 0777 );
	}

	if( ftp->LocalDir( locdir, LocDirList ) == -1 )
	{
		if (warnings >= 0)
			fprintf( stderr, "Error reading local directory `%s'!\n", locdir );
		RemDirList.Destroy();
		return -1;
	}

	printf ("%s:\n", dir);

	FOR_CWLIST( RemDirList, RemNode )
	{
		if( RemNode->FDN_Name != NULL
		 && RemNode->FDN_Mode != 0
		 && RemNode->FDN_Name[0] != '.' )
		{
			sprintf( locfile, "%s/%s", locdir, RemNode->FDN_Name );
			sprintf( remfile, "%s/%s", remdir, RemNode->FDN_Name );
			sprintf( oldfile, "%s/.old.%s", locdir, RemNode->FDN_Name );
			sprintf( newfile, "%s/.new.%s", locdir, RemNode->FDN_Name );

			ret = stat( locfile, &sbuf );	/* get info on file */
			sterr = errno;
			getit = 0;			/* do not get it by default */
			reget = 0;			/* do not restart by default */
			time( &utbuf.actime );		/* access time */
			utbuf.modtime = RemNode->FDN_Date;	/* mod time */

			tfs.total++;
			if (verbose > 0) printf( "%s:", RemNode->MakeStr() );

			asascii = AsciiPattern.Match (locfile);
			tfs.f_ascii+=asascii;

			if( IncludePattern.Pattern() != NULL
			 && !IncludePattern.Match( RemNode->FDN_Name ) )
			{
				if (verbose > 0) printf( " not included\n" );
				tfs.f_ignored++;
			}
			else if( ExcludePattern.Match( RemNode->FDN_Name ) )
			{
				if (verbose > 0) printf( " excluded\n" );
				tfs.f_ignored++;
			}
			else if( ( RemNode->FDN_Mode & S_IFDIR ) != 0 )
			{
				if (verbose > 0) printf( " subdir\n" );
				tfs.dirs++;
				if( subdirs )
				{
					makedirstr (s, dir, RemNode->FDN_Name);
					SubDirList.Append (new MirrorSubDir(s));
				}
			}
			else if( ( RemNode->FDN_Mode & S_IFREG ) != 0 )
			{
				if( ret != 0 ) /* ret is from stat() */
				{
					if (verbose > 0) printf( " new file\n" );
					getit = 1;
					tfs.f_new++;
				}
				else if( RemNode->FDN_Date > sbuf.st_mtime && newfiles == 0 )
				{
					if (verbose > 0) printf( " new date (diff=%ld)\n",
						(RemNode->FDN_Date-sbuf.st_mtime) );
					getit = 1;
					tfs.f_mod++;
				}
				// ignore size difference if transferring ascii
				else if( !asascii && RemNode->FDN_Size != sbuf.st_size && newfiles == 0 )
				{
					if( RemNode->FDN_Size > sbuf.st_size )
					{
						reget = 1; /* local file is smaller, do reget */
						if (verbose > 0) printf( " reget" );
					}
					if (verbose > 0) printf( " new size (old %ld)\n", (long) sbuf.st_size );
					getit = 1;
					tfs.f_mod++;
				}
				else if( force )
				{
					if (verbose > 0) printf( " get forced\n" );
					getit = 1;
				}
				else
				{
					tfs.f_same++;
					if (verbose > 0) printf( " same\n" );
				}

				if( getit )
				{
					/*
					 * Get the exact time of the remote file
					 */

					if( ( tmpt = ftp->FileTime( remfile ) ) == -1 )
					{
						if (warnings >= 1)
							fprintf( stderr, "Warning: Can't get time of remote"
								" file %s, \n\tusing directory information\n",
								remfile );
					}
					else
					{
						RemNode->FDN_Date = tmpt;
					}

					/*
					 * Check whether there is an old newfile from the last
					 * transfer
					 */

					newfilelength = 0;

					if( stat( newfile, &newfilestat ) == 0 )
					{
						newfilelength = newfilestat.st_size;
					}

					if( ftp->Get(
						remfile,
						reget ? locfile : newfile,
						RemNode->FDN_Size,
						asascii) )
					{
						fprintf( stderr, "Error retrieving remote file `%s'!\n", remfile );
						ret = stat( reget ? locfile : newfile, &sbuf );
					}
					else
					{
						tfs.f_got++;
						if( ret == 0 )
						{
							if( saveold )
							{
							    remove( oldfile );
							    rename( locfile, oldfile );
							}
							else
							{
								remove( locfile );
							}
						}

						if( reget == 0 ) rename( newfile, locfile );
						utime( locfile, &utbuf );
						ret = stat( locfile, &sbuf );
					}
				}
			}
			else
			{
				if (verbose > 0) printf( " unknown type\n" );
			}

			fflush( stdout );
		}
	}

	if( !nodelete )
	{
		FOR_CWLIST( LocDirList, LocNode )
		{
			if (verbose > 1) printf ("%s:", LocNode->FDN_Name);
			FOR_CWLIST( RemDirList, RemNode )
			{
				if( LocNode->FDN_Name != NULL
				 && RemNode->FDN_Name != NULL
				 && strcmp( LocNode->FDN_Name, RemNode->FDN_Name ) == 0 ) break;
			}

			if( RemNode == NULL ) /* Not found */
			{
				sprintf( locfile, "%s/%s", locdir, LocNode->FDN_Name );
				sprintf( delfile, "%s/.del.%s", locdir, LocNode->FDN_Name );

				// keep files not on remote anyway
				if( KeepPattern.Match( LocNode->FDN_Name ) ) {
					tfs.f_kept++;
					if (verbose > 0) printf (" kept anyway.\n");
					continue;
				}

				tfs.f_rem++;
				if( deloldf )
				{
					if (verbose > 0) printf( " deleted oldfile.\n" );
					remove( locfile );
				}
				else
				{
					if (verbose > 0) printf( " renamed oldfile.\n" );
					remove( delfile );
					rename( locfile, delfile );
				}
			}
			else 
				if (verbose > 0) printf (" kept.\n");
		}
	}

	RemDirList.Destroy();
	LocDirList.Destroy();

	tfs.print(dir);

	FOR_CWLIST( SubDirList, subdir ) {
		mirror( ftp, localdir, remotedir, subdir->dir, newfiles );
	}
	SubDirList.Destroy();

	return 0;
}

/* ---------------------------------------------------------------------- */

char *program = "cwftp";

char hostname[256] = "localhost";
char username[256] = "ftp";
char password[256] = "mirror@";
char localdir[256];
char remotedir[256] = "/";

void usage()
{
	printf( "cwftp version " VERSION ", Copyright (C) 1995 Christian Wolf\n" );
	printf( "cwftp comes with ABSOLUTELY NO WARRANTY\n" );
	printf( "This is free software, and you are welcome to redistribute it\n" );
	printf( "under certain conditions; see COPYING for details.\n" );
	printf( "\n" );
	printf( "usage: %s [options] directory ...\n\n", program );
	printf( "  -u username   loginname (%s)\n", username );
	printf( "  -p password   loginpassword (%s)\n", password );
	printf( "  -h hostname   remote hostname (%s)\n", hostname );
	printf( "  -l localdir   localdir (%s)\n", localdir );
	printf( "  -r remotedir  remotedir (%s)\n", remotedir );
	printf( "  -o            save old files\n" );
	printf( "  -n            get only new files\n" );
	printf( "  -s            get subdirectories\n" );
	printf( "  -d            delete obsolete files\n" );
	printf( "  -v            be verbose\n" );
	printf( "  -w            print warnings\n" );
	printf( "  -A ascii      ascii-pattern (%s)\n",
		AsciiPattern.Pattern() ? AsciiPattern.Pattern() : "not set" );
	printf( "  -E exclude    exclude-pattern (%s)\n",
		ExcludePattern.Pattern() ? ExcludePattern.Pattern() : "not set" );
	printf( "  -I include    include-pattern (%s)\n",
		IncludePattern.Pattern() ? IncludePattern.Pattern() : "not set" );
	printf( "  -K keep       keep-pattern (%s)\n",
		KeepPattern.Pattern() ? KeepPattern.Pattern() : "not set" );

	printf( "\n" );
}

int main( int argc, char **argv )
{
	char myhost[256];
	char mydomain[256];
	char mylogin[256];
	int opt;
	tFTP *ftp;
	char *p;

	program = argv[0];
	gethostname( myhost, sizeof(myhost) );
	getdomainname( mydomain, sizeof(mydomain) );
	p = getlogin();
	if (p)
		strcpy( mylogin, p ); // getlogin fails sometimes when running from cron(?)
	else
		strcpy (mylogin, "nobody");

	getcwd( localdir, sizeof(localdir) );
	sprintf( password, "%s@", mylogin );

	while( ( opt = getopt( argc, argv, "de:fh:i:l:noOp:r:stu:vwA:E:I:K:" ) ) != -1 )
	{
		switch( opt )
		{
		case 'd':
			deloldf++;
			break;

		case 'f':
			force++;
			break;

		case 'h':
			strcpy( hostname, optarg );
			break;

		case 'l':
			strcpy( localdir, optarg );
			break;

		case 'n':
			newonly++;
			break;

		case 'o':
			saveold++;
			break;

		case 'O':
			nodelete++;
			break;

		case 'p':
			strcpy( password, optarg );
			break;

		case 'r':
			strcpy( remotedir, optarg );
			break;

		case 's':
			subdirs = 1;
			break;

		case 't':
			do_transmit = 1;
			break;

		case 'u':
			strcpy( username, optarg );
			break;

		case 'v':
			verbose++;
			break;
		
		case 'w':
			warnings++;
			break;

		case 'A':
			AsciiPattern.SetNewExpression( optarg );
			break;

		case 'e':
		case 'E':
			ExcludePattern.SetNewExpression( optarg );
			break;

		case 'i':
		case 'I':
			IncludePattern.SetNewExpression( optarg );
			break;

		case 'K':
			KeepPattern.SetNewExpression( optarg );
			break;

		default:
			printf ("Unknown option `%c'\n", opt);
			usage();
			exit( 1 );
		}
	}

	if( optind == argc )
	{
		usage();
		exit( 1 );
	}

	ftp = new tFTP( verbose, warnings );

	printf( "connect to %s with login=%s password=%s ...\n",
		hostname, username, password );

	if( ftp->Open(
		hostname,
		username,
		password ) )
	{
		fprintf( stderr, "Error connecting to `%s'!\n", hostname );
		exit( 1 );
	}

	while( optind < argc )
	{
		if( do_transmit )
		{
			transmit( ftp, localdir, remotedir, argv[optind] );
		}
		else
		{
			mirror( ftp, localdir, remotedir, argv[optind], newonly );
		}
		optind++;
	}

	ftp->Close();

	delete ftp;

	return 0;
}
