/*
 * $Id: CWList.H,v 1.1 1996/01/30 09:26:04 chris Exp $
 *
 * $Log: CWList.H,v $
 * Revision 1.1  1996/01/30  09:26:04  chris
 * Initial revision
 *
 *
 * $Desc: template List class and extern iterator (header+implementation)$
 */

#ifndef CWLIST_H
#define CWLIST_H

#include <stdlib.h> /* fuer NULL */

template<class Item> class tCWIterator;

template<class Item> class tCWList
{
	friend class tCWIterator<Item>;

private:
	struct tCWListNode
	{
		tCWListNode *succ;
		tCWListNode *pred;
		Item        *obj;
	};

	tCWListNode *head;
	tCWListNode *tail;
	tCWListNode *cursor;
	int len;

	tCWListNode *SearchNode( Item *search )
	{
		tCWListNode *node;

		for( node=head; node != NULL; node = node ->succ )
		{
			if( node->obj == search ) return node;
		}
		return NULL;
	}

public:
	/* Liste leeren */

	void DeleteAll()
	{
		while( ( cursor = head) != NULL )
		{
			head = head->succ;
			delete cursor;
		}
		tail = NULL;
		len = 0;
	}

	void Destroy()
	{
		while( ( cursor = head) != NULL )
		{
			head = head->succ;
			delete cursor->obj;
			delete cursor;
		}
		tail = NULL;
		len = 0;
	}

	tCWList()
	{
		head = tail = cursor = NULL;
		len = 0;
	}

	~tCWList()
	{
		DeleteAll();
	}

	/* Listen verwaltung */

	void Append( Item *Obj )
	{
		tCWListNode *node = new tCWListNode;

		node->succ = NULL;
		node->pred = tail;
		node->obj  = Obj;

		if( tail == NULL ) /* First Node in this List */
		{
			head = node;
		}
		else
		{
			tail->succ = node;
		}
		tail = node;
		len++;
	}

	int Delete( Item *Obj = NULL )
	{
		tCWListNode *node;

		if( Obj == NULL )
		{
			node = cursor;
		}
		else
		{
			node = SearchNode( Obj );
		}

		if( node != NULL )
		{
			if( node->succ == NULL )
			{
				tail = node->pred;
			}
			else
			{
				(node->succ)->pred = node->pred;
			}

			if( node->pred == NULL )
			{
				head = node->succ;
			}
			else
			{
				(node->pred)->succ = node->succ;
			}

			delete node;
			len--;

			return 0;
		}
		else
		{
			return -1;
		}
	}

	int Test( Item *Obj )
	{
		return ( SearchNode( Obj ) != NULL ) ? 1 : 0;
	}

	/* Iterator 1 */

	Item *GetFirst()
	{
		cursor = head;
		if( cursor != NULL ) return cursor->obj;
		return NULL;
	}

	Item *GetNext()
	{
		if( cursor != NULL ) cursor = cursor->succ;

		if( cursor != NULL ) return cursor->obj;
		return NULL;
	}

	/* Iterator 2 */

	void Head()
	{
		cursor = head;
	}

	void Tail()
	{
		cursor = tail;
	}

	void Prev()
	{
		if( cursor != NULL ) cursor = cursor->pred;
	}

	void Next()
	{
		if( cursor != NULL ) cursor = cursor->succ;
	}

	Item *Get()
	{
		if( cursor != NULL ) return cursor->obj;
		else return NULL;
	}

	int isHead()
	{
		return ( head != NULL && cursor == head );
	}

	int isTail()
	{
		return ( tail != NULL && cursor == tail );
	}

	int isEmpty()
	{
		return ( head != NULL );
	}

	int Len()
	{
		return len;
	}
};

template<class Item> class tCWIterator
{
private:
	tCWListNode *Cursor;
	tCWList<Item> *List;
public:
	tCWIterator( tCWList<Item> *aList )
	{
		List = aList;
		Cursor = List->head;
	}

	void SetList( tCWList<Item> *aList )
	{
		List = aList;
		Cursor = List->head;
	}

	/* Iterator 1 */

	Item *GetFirst()
	{
		Cursor = List->head;
		if( Cursor != NULL ) return Cursor->obj;
		return NULL;
	}

	Item *GetNext()
	{
		if( Cursor != NULL ) Cursor = Cursor->succ;
		if( Cursor != NULL ) return Cursor->obj;
		return NULL;
	}
#if 0
	/* Iterator 2 */

	void Head()
	{
		Cursor = List->head;
	}

	void Tail()
	{
		Cursor = List->tail;
	}

	void Prev()
	{
		if( Cursor != NULL ) Cursor = Cursor->pred;
	}

	void Next()
	{
		if( Cursor != NULL ) Cursor = Cursor->succ;
	}

	Item *Get()
	{
		if( Cursor != NULL ) return Cursor->obj;
		else return NULL;
	}

	int isHead()
	{
		return ( List->head != NULL && Cursor == List->head );
	}

	int isTail()
	{
		return ( List->tail != NULL && Cursor == List->tail );
	}
#endif
};

#define FOR_CWLIST( l, p ) for((p)=(l).GetFirst();(p)!=NULL;(p)=(l).GetNext())
#define FOR_CWITER( i, p ) for((p)=(i).GetFirst();(p)!=NULL;(p)=(i).GetNext())

#endif /* CWLIST_H */
