/*
 *  Ftp.C - Generic FTP C++ class (implementation)
 *  Copyright (C) 1995 Christian Wolf
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *  You can reach me, the author of cwftp as <chris@MedIS.DE>
 *
 * $Id: Ftp.C,v 1.3 1996/05/13 16:49:54 chris Exp $
 *
 * $Log: Ftp.C,v $
 * Revision 1.3  1996/05/13 16:49:54  chris
 * STOR, Bugfixes
 *
 * Revision 1.2  1996/01/30  09:26:04  chris
 * *** empty log message ***
 *
 * Revision 1.1  1995/05/08  11:47:52  chris
 * Initial revision
 *
 *
 */

#include <stdarg.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/stat.h>
/* Ftp.H: #include <netinet/in.h> */
/* Ftp.H: #include <stdio.h> */
#include <ctype.h>
#include <string.h>
#include <netdb.h>
#include <unistd.h>
#include <arpa/ftp.h>
#include <arpa/telnet.h>
#include <errno.h>
#include <fcntl.h>
/* Ftp.H: #include <time.h> */
#include <utime.h>

/* #include <linux/in.h> */
#include <arpa/nameser.h>
/* #include <resolv.h> */

#ifdef HAVE_DIRENT
#include <dirent.h>
typedef struct dirent d_type;
#define D_NAME(d) ((d)->d_name)
#else
#include <sys/dir.h>
typedef struct direct d_type;
#define D_NAME(d) ((d)->d_name)
#endif

#include "Ftp.H"
#include "Split.H"
#include "TimeVal.H"

#ifdef __svr4__
#define bcopy( s, d, l ) memcpy( d, s, l )
#define bzero( d, l ) memset( d, 0, l );
#endif

#define LSTYPE_UNKNOWN	0
#define LSTYPE_UNIX	1
#define LSTYPE_HP	2
#define LSTYPE_MACOS	3

/*
** Sample for Unix ls output:
**
** total 2516
** -rw-r--r--  1 chris    wolf       729260 Jan 15  1977 FSUU12R4.ZIP
** ...
** -rw-r--r--  1 chris    729260 Jan 15  1977 FSUU12R4.ZIP
**
**
** Sample for HP/RTE Output:
**
** directory /RUN.DIR 
**        name        ex     prot   type blks  words recs   addr/lu
** ABIOCHANGESTAT.RUN       rw/rw/rw      6   73   9344   73  25004/27
** ...
** directory ::RUN 
**          name        ex     prot type blks words recs  addr/lu
**
** ABIOINFO.RUN            rw/rw/rw    6   97 12416   97 25504/27
** ...
** WOCHENSAVE.FST        * rw/rw/rw    4    4   433   36  1464/27
**
**
** Sample for MacOS output: (NetPresenz, formerly FTPd by Peter N. Lewis)
**
** ----r-----           0      555      555 Apr 27  1994 desc.txt
** drwx------               folder        0 Jan 24 14:13 u0072
*/

struct SystemType { char *name; int lstype;
};

struct SystemType SystemTypes[] = {
	{ "UNIX",	LSTYPE_UNIX },
	{ "MacOS",	LSTYPE_MACOS },
	{ 0, 0 }
};

inline char *mydup( char *ptr )
{
	char *np = new char[ strlen(ptr) + 1 ];
	strcpy( np, ptr );
	return np;
}

static char *ntosl (unsigned long a) {
	static char s[256];

	a = ntohl (a);
	sprintf (s, "%lu.%lu.%lu.%lu", 
		(a>>24&0xFF), 
		(a>>16&0xFF), 
		(a>>8&0xFF), 
		a&0xFF);
	return s;
}

tFtpDirNode::tFtpDirNode()
{
	FDN_Name = NULL;
	FDN_Size = -1L;
	FDN_Mode = 0L;
	FDN_Date = 0L;
}

tFtpDirNode::~tFtpDirNode()
{
	if( FDN_Name != NULL ) delete FDN_Name;
}

char *tFtpDirNode::MakeStr()
{
	static char Modes[] = "-rwxrwxrwx";
	static char txt[80];

	struct tm *GMT;
	int pos;

	GMT = gmtime( &FDN_Date );

	sprintf( txt, "?--------- %8ld %02d.%02d.%02d %02d:%02d:%02d %s",
		FDN_Size,
		GMT->tm_mday,
		GMT->tm_mon+1,
		GMT->tm_year,
		GMT->tm_hour,
		GMT->tm_min,
		GMT->tm_sec,
		FDN_Name );

	switch( FDN_Mode & S_IFMT ) /* 0xf000 */
	{
	case S_IFIFO: /* 0x1000 */
		txt[0] = 'f';
		break;
	case S_IFCHR: /* 0x2000 */
		txt[0] = 'c';
		break;
	case S_IFDIR: /* 0x4000 */
		txt[0] = 'd';
		break;
	case S_IFBLK: /* 0x6000 */
		txt[0] = 'b';
		break;
	case S_IFREG: /* 0x8000 */
		txt[0] = '-';
		break;
	case S_IFLNK: /* 0xA000 */
		txt[0] = 'l';
		break;
	case S_IFSOCK: /* 0xC000 */
		txt[0] = 's';
		break;
	default:
		txt[0] = '?';
		break;
	}

	for( pos=8; pos>=0; pos-- )
	{
		if( ( FDN_Mode & (1<<pos) ) != 0 )
		{
			txt[9-pos] = Modes[9-pos];
		}
	}

	if( ( FDN_Mode & S_ISUID ) != 0 ) txt[3] = 's';
	if( ( FDN_Mode & S_ISGID ) != 0 ) txt[6] = 's';
	if( ( FDN_Mode & S_ISVTX ) != 0 ) txt[9] = 't';

	return txt;
}

tFTP::tFTP( int aVerbose, int aWarnings )
{
	Verbose  = aVerbose;
	Warnings = aWarnings;

	DirLen = 0;
	DirPos = 0;
	DirSiz = 0;
	DirBuf = NULL;

	CommandSD = -1;
	DataSD = -1;
	BindSD = -1;
	Connected = 0;
	Code = 0;
	InLen = 0;
	InPos = 0;
	UsePASV = 1;
	
	LsType = LSTYPE_UNKNOWN;
};

tFTP::~tFTP()
{
};

void tFTP::LostPeer()
{
	if( Connected != 0 )
	{
		if( CommandSD >= 0 )
		{
			(void) shutdown( CommandSD, 2 );
			(void) close( CommandSD );
			CommandSD = -1;
		}
		if( DataSD >= 0 )
		{
			(void) shutdown( DataSD, 2 );
			(void) close( DataSD );
			DataSD = -1;
		}
		Connected = 0;
		if( Verbose > 0 ) puts( "*** Lost peer" );
	}
	InLen = 0;
	InPos = 0;
}

void tFTP::CloseData()
{
	if( BindSD >= 0 )
	{
		shutdown( BindSD, 2 );
		close( BindSD );
		BindSD = -1;
	}
	if( DataSD >= 0 )
	{
		shutdown( BindSD, 2 );
		close( DataSD );
		DataSD = -1;
	}
}

int tFTP::GetC()
{
	if( InPos == InLen )
	{
		InLen = read( CommandSD, InBuf, sizeof(InBuf ) );
		InPos = 0;
	}

	if( InPos < InLen )
	{
		return InBuf[ InPos++ ];
	}
	return EOF;
}

int tFTP::GetReply( int waiteof )
{
	int ch;
	int FirstDigit;
	int Digit;
	int FirstCode = 0;
	int Continue = 0;
	char tbuf[4];
	int rpos = 0;

	for (;;)
	{
		/*
		 * Zeilenanfang
		 */

		Digit = 0;
		FirstDigit = 0;
		Code = 0;
		ReplyBuf[0] = '\0';

		while( ( ch = GetC() ) != '\n' )
		{
			if( ch == IAC )
			{
				/* Telnet */
				switch( ch = GetC() )
				{
				case WILL:
				case WONT:
					ch = GetC();
					sprintf( tbuf, "%c%c%c", IAC, DONT, ch );
					write( CommandSD, tbuf, 3 );
					break;
				case DO:
				case DONT:
					ch = GetC();
					sprintf( tbuf, "%c%c%c", IAC, WONT, ch );
					write( CommandSD, tbuf, 3 );
					break;
				default:
					break;
				}
				continue;
			}

			Digit++;

			if( ch == EOF )
			{
				if( Verbose > 2 ) printf( "<-- Connections closed\n" );

				if( waiteof )
				{
					Code = 221;
					return 0;
				}
				Code = 421;
				LostPeer();
				return 4;
			}
			if( ch != '\r' ) ReplyBuf[ rpos++ ] = ch;

			if( Digit < 4 && isdigit(ch) ) Code = Code * 10 + (ch - '0');

			if( Digit == 4 && ch == '-' )
			{
				if( Continue ) Code = 0;
				Continue++;
			}
			if( FirstDigit == 0 ) FirstDigit = ch;
		}

		/*
		 * Eine Zeile ist voll
		 */

		ReplyBuf[rpos]='\0';
		rpos = 0;

		if( Verbose > 2 ) printf( "<-- %s\n", ReplyBuf );

		if( Continue && Code != FirstCode )
		{
			if( FirstCode == 0 ) FirstCode = Code;
			continue;
		}

		if( Code == 421 || FirstCode == 421 ) LostPeer();

		return ( FirstDigit - '0');
	}
}

void tFTP::SendCommand ( char *fmt, ... )
{
	va_list vl;
	char cmd[BUFSIZ];

	va_start( vl, fmt );
	vsprintf( cmd, fmt, vl);
	va_end( vl );

	if( Verbose > 2 ) printf( "--> %s\n", cmd );

	strcat( cmd, "\r\n" );
	write( CommandSD, cmd, strlen( cmd ) );
}


int tFTP::Command( char *fmt, ... )
{
	va_list vl;
	char cmd[BUFSIZ];

	va_start( vl, fmt );
	vsprintf( cmd, fmt, vl);
	va_end( vl );

	if( Verbose > 2 ) printf( "--> %s\n", cmd );

	strcat( cmd, "\r\n" );
	write( CommandSD, cmd, strlen( cmd ) );

	return GetReply( strncmp( cmd, "QUIT", 4 ) == 0 );
}

// send command and open data connection
int tFTP::CommandData( char *fmt, ... )
{
	va_list vl;
	char cmd[BUFSIZ];
	int n;

	va_start( vl, fmt );
	vsprintf( cmd, fmt, vl);
	va_end( vl );

	if (UsePASV) {
		SendCommand ("%s", cmd);
	} else {
		n = Command("%s", cmd);
		if (n != PRELIM ) {
			CloseData();
			return -1;
		}
	}
	
	if( DataConnection() == -1 ) {
		Code = -1;
		CloseData();
		return -1;
	}
	
	if (UsePASV) {
		n = GetReply(0);
		if (n != PRELIM) {
			CloseData();
			return -1;
		}
	}
	return 0;
}

char *tFTP::HErrorStr( int err )
{
	switch( err )
	{
	case HOST_NOT_FOUND: return "Authoritive Answer Host not found";
	case TRY_AGAIN: return "Non-Authoritive Host not found, or SERVERFAIL";
	case NO_RECOVERY: return "Non recoverable errors, FORMERR, REFUSED, NOTIMP";
	case NO_DATA: return "Valid name, no data record of requested type";
	}
	return "unknown error";
}

/*
 * DataConnection annehmen(PORT) oder aufbauen(PASV)
 */

int tFTP::DataConnection()
{
	struct sockaddr_in from;
	int fromlen = sizeof (from);

	if( DataSD >= 0 )
	{
		close( DataSD );
		DataSD = -1;
	}

	if( UsePASV ) {
		if( ( DataSD = socket( AF_INET, SOCK_STREAM, 0 ) ) == -1 )
		{
			perror( "socket" );
			CloseData();
			return -1;
		}

		if (Verbose > 1) {
			printf ("Opening data connection to %s:%u\n", 
				ntosl (DataAddress.sin_addr.s_addr), 
				ntohs (DataAddress.sin_port));
		}

		DataAddress.sin_family = AF_INET;

		if( connect( DataSD, (struct sockaddr *)&DataAddress,
			sizeof(DataAddress) ) == -1 )
		{
			perror( "connect" );
			CloseData();
			return -1;
		}
	}
	else
	{
		if (Verbose > 1) {
			printf ("Waiting for data connection on %s:%u\n", 
				ntosl (DataAddress.sin_addr.s_addr), 
				ntohs (DataAddress.sin_port));
		}

		DataSD = accept( BindSD, (struct sockaddr *) &from, &fromlen );

		if( DataSD == -1 )
		{
			perror( "accept" );
			CloseData();
			return -1;
		}

#ifdef CLOSE_BIND
		close( BindSD );
		BindSD = -1;
#endif
	}

	if( Verbose > 2 ) printf( "DataConnection open\n" );
	return DataSD;
}

int tFTP::InitConnection()
{
	register unsigned char *p, *a;
	int result, len;
	char *portstr;

	a = (unsigned char *)&DataAddress.sin_addr;
	p = (unsigned char *)&DataAddress.sin_port;

	if( UsePASV )
	{
		if( Command( "PASV" ) != 2 )
		{
			if (Warnings >= 1)
			fprintf( stderr, "Warning: remote server does not support PASV, using PORT.\n" );
			UsePASV = 0;
		}
		else
		{
			unsigned int Val[6];

			if( ( portstr = strchr( ReplyBuf, '(' ) )
				&& sscanf( portstr, "(%u,%u,%u,%u,%u,%u)",
					Val, Val+1, Val+2, Val+3, Val+4, Val+5 ) == 6 )
			{
				DataAddress.sin_addr.s_addr = 
					htonl (Val[0]<<24 | Val[1]<<16 | Val[2]<<8 | Val[3]);
				DataAddress.sin_port = 
					htons (Val[4]<<8 | Val[5]);

				return 0;
			}
			else
			{
				if (Warnings >= 1)
					fprintf( stderr, "Warning: Can't read PASV-reply `%s', using PORT.\n", ReplyBuf );
				UsePASV = 0;
			}
		}
	}

	if( UsePASV == 0 )
	{
		DataAddress = MyClientAddress;
		DataAddress.sin_port = 0;	/* let system pick one */

#ifdef CLOSE_BIND
		if( BindSD >= 0 )
		{
			close( BindSD );
			BindSD = -1;
		}
#endif
		if( BindSD < 0 )
		{
			if( ( BindSD = socket(AF_INET, SOCK_STREAM, 0) ) < 0 )
			{
				perror( "socket" );
				return -1;
			}
		}

		if( bind( BindSD, (struct sockaddr *)&DataAddress, sizeof (DataAddress)) < 0)
		{
			perror( "bind" );
			close( BindSD );
			BindSD = -1;
			return -1;
		}

		len = sizeof (DataAddress);
		if( getsockname( BindSD, (struct sockaddr *)&DataAddress, &len) < 0)
		{
			perror( "getsockname" );
			close( BindSD );
			BindSD = -1;
			return -1;
		}

		listen( BindSD, 1);

#define	UC(b)	(((int)b)&0xff)

		result = Command( "PORT %d,%d,%d,%d,%d,%d",
			UC(a[0]), UC(a[1]), UC(a[2]), UC(a[3]), UC(p[0]), UC(p[1]));

		if( result == COMPLETE) return 0;
	}

	return -1;
}

int tFTP::GetSystemType () {
	int n;
	struct SystemType *p;
	
	n = Command ("SYST");
	if (n == 2) {
		for (p = SystemTypes; p->name; p++) {
			if (strncasecmp (p->name, ReplyBuf+4, strlen (p->name)) == 0) {
				LsType = p->lstype;
				if (Verbose > 0)
					printf ("Remote system type is %s\n", p->name);
				return 0;
			}
		}
	}
	if (Verbose > 0)
		printf ("Remote system type could not be determined.\n\t(%s)\n", ReplyBuf+4);
	return 1;
}

int tFTP::Open( char *hostname, char *username, char *password )
{
	struct sockaddr_in server;
	struct servent *sp;
	struct hostent *hp;
	int len,n;

	if( Connected )
	{
		Command("QUIT");
		if( CommandSD != -1 ) close( CommandSD );
		CommandSD = -1;
		Connected = 0;
		BindSD = -1;
	}

	if( ( sp = getservbyname( "ftp","tcp" ) ) == NULL)
	{
		perror( "getservbyname" );
		return -1;
	}

	if( ( hp = gethostbyname( hostname ) ) == NULL)
	{
		if (Warnings >= 0)
			fprintf( stderr, "Error: can't get name of server `%s': %s\n",
				hostname, HErrorStr( h_errno ) );
		return -1;
	}

	bzero((char *)&server,sizeof(server));
	bcopy(hp->h_addr,(char *)&server.sin_addr,hp->h_length);
	server.sin_family = hp -> h_addrtype;
	server.sin_port = sp ->s_port;

	if( CommandSD >= 0 )
	{
		close( CommandSD );
		CommandSD = -1;
	}

	if( ( CommandSD = socket(AF_INET,SOCK_STREAM,0) ) < 0 )
	{
		perror( "socket" );
		return -1;
	}

	if( Verbose > 1 )
	{
		unsigned char *a = (unsigned char *)&(server.sin_addr.s_addr);
		printf( "Trying %u.%u.%u.%u ...\n",
			a[0], a[1], a[2], a[3] );
	}

	if( connect( CommandSD, (struct sockaddr *)&server, sizeof(server) ) < 0 )
	{
		perror( "connect" );
		return -1;
	}

	if( Verbose > 0 ) printf( "Connection to %s established\n", hostname );

	len = sizeof (MyClientAddress);
	if (getsockname( CommandSD,(struct sockaddr *)&MyClientAddress, &len) <0 )
	{
		perror( "getsockname" );
		return -1;
	}
	MyClientAddress.sin_family = server.sin_family;

	Connected = 1;

	/*
	 * Login-Meldung vom FTP-Server abwarten
	 */

	if( GetReply(0) > 2 )
	{
		LostPeer();
		return -1;
	}

	/* ftp-login */

	n = Command( "USER %s", username );
	if (n == CONTINUE) n = Command("PASS %s", password );
	if (n == CONTINUE) n = Command("ACCT %s", password );
	if (n != COMPLETE)
	{
		LostPeer();
		return -1;
	}

	printf( "Login to %s successful.\n", hostname );
	
	GetSystemType();

	return 0;
}

int tFTP::Close()
{
	LostPeer();
	Code = 0;
	return 0;
}

int tFTP::ChangeDir( char *dirname )
{
	if( Command( "CWD %s", dirname ) != COMPLETE )
	{
		return -1;
	}
	return 0;
}

// remove CRs
static int AsciiConvert (char *d, const char *s, int len) {
	int newlen = 0;
	
	while (len--) {
		if (*s != 13) {
			*d++ = *s++;
			newlen++;
		}
		else
			s++;
	}
	return newlen;
}

int tFTP::Get( char *filename, char *localname, long aFileSize, int asAscii )
{
	char buffer[8192], buffer2[8192];
	long bytes = 0;
	int d = 0, c;
	int fd;
	int do_append = 0;
	TimeVal t1, t2;
	double cps;
	struct stat oldfilestat;
	long oldfilesize = 0;

	// don't do reget on ascii
	if( !asAscii && stat( localname, &oldfilestat ) == 0 )
	{
		oldfilesize = oldfilestat.st_size;
	}

	if( Command(asAscii ? "TYPE A" : "TYPE I")!=COMPLETE)
	{
		LostPeer();
		return -1;
	}

	if( InitConnection() )
	{
		Code = -1;
		return -1;
	}

	/*
	 * Restart at old Position
	 */

	if( oldfilesize > 0 )
	{
		do_append = 1;
		if (Verbose > 0) printf( "Restart RETR of %s at %ld\n", filename, oldfilesize );
		Command( "REST %ld", oldfilesize );
	}

	if( CommandData( "RETR %s",filename) != PRELIM )
		return -1;
	if( oldfilesize > 0 )
	{
		fd = open( localname, O_APPEND | O_WRONLY );
	}
	else
	{
		fd = open( localname, O_CREAT|O_WRONLY, 0644 );
	}

	if( fd == -1 )
	{
		perror( localname );
		Code = -1;
		CloseData();
		return -1;
	}

	t1.SetNow();

	while( ( c = read( DataSD, buffer, sizeof(buffer) ) ) > 0 )
	{
		if (asAscii) {
			c = AsciiConvert (buffer2, buffer, c);
			if( ( d = write( fd, buffer2, c ) ) != c) break;
		} else {
			if( ( d = write( fd, buffer, c ) ) != c) break;
		}
		bytes += c;

		if( Verbose > 3 || isatty (0))
		{
			if( aFileSize > 0 )
			{
				t2 = TimeVal::Now() - t1;
				if( t2.GetSeconds() == 0 && t2.GetMicros() == 0 )
					cps = 1.0;
				else
					cps = ((double)bytes) * 1000000.0 / 
						(1000000 * t2.GetSeconds() + t2.GetMicros());

				printf( "%s %0.1fK/%0.1fK %ld%% %0.1fK/s         \r",
					filename, (oldfilesize+ bytes)/1024.0, aFileSize/1024.0,
					(oldfilesize+bytes)*100/aFileSize, cps/1024 );
			}
			else
			{
				printf( "%s %ld\r", filename, oldfilesize+bytes );
			}
			fflush( stdout );
		}
	}

	t2 = TimeVal::Now() - t1;

	if( Verbose > 0 )
	{
		if( t2.GetSeconds() == 0
		 && t2.GetMicros() == 0 )
		{
			cps = 1.0;
		}
		else
		{
			cps = 1000000 * t2.GetSeconds() + t2.GetMicros();
			cps = ((double)bytes) * 1000000.0 / cps;
		}

		printf( "### %s: got %ld bytes in %ld.%06ld sec. (%ld cps)\n",
			filename, bytes, t2.GetSeconds(), t2.GetMicros(), (long)cps );
	}

	if (d < c) perror( "write" );

	close( fd );

	CloseData();

	/*
	 * Antwort vom FTP-Server: "226 Transfer complete."
	 */

	if( GetReply( 0 ) != 2 )
	{
		if (Warnings >= 0)
			fprintf( stderr, "Error retrieving file `%s': %s\n",
				filename,  ReplyBuf);
		return -1;
	}

	return 0;
}

int tFTP::Put( char *filename, char *localname, int asAscii )
{
	char buffer[8192], buffer2[8192];
	long bytes = 0;
	int d = 0, c;
	int fd;
	TimeVal t1, t2;
	double cps;
	struct stat locfilestat;
	long locfilesize = 0;
	long oldfilesize = 0; /* reput is a TODO */

	if( stat( localname, &locfilestat ) == -1 )
	{
		perror( localname );
		return -1;
	}
	locfilesize = locfilestat.st_size;

	if( Command(asAscii ? "TYPE A" : "TYPE I")!=COMPLETE)
	{
		LostPeer();
		return -1;
	}

	if( InitConnection() )
	{
		Code = -1;
		return -1;
	}

	if( CommandData( "STOR %s",filename) != PRELIM )
		return -1;

	fd = open( localname, O_RDONLY );

	if( fd == -1 )
	{
		perror( localname );
		Code = -1;
		CloseData();
		return -1;
	}

	t1.SetNow();

	while( ( c = read( fd, buffer, sizeof(buffer) ) ) > 0 )
	{
		if (asAscii) {
			c = AsciiConvert (buffer2, buffer, c);
			if( ( d = write( DataSD, buffer2, c ) ) != c) break;
		} else {
			if( ( d = write( DataSD, buffer, c ) ) != c) break;
		}
		bytes += c;

		if( Verbose > 3 || isatty (0))
		{
			if( locfilesize > 0 )
			{
				t2 = TimeVal::Now() - t1;
				if( t2.GetSeconds() == 0 && t2.GetMicros() == 0 )
					cps = 1.0;
				else
					cps = ((double)bytes) * 1000000.0 / 
						(1000000 * t2.GetSeconds() + t2.GetMicros());

				printf( "%s %0.1fK/%0.1fK %ld%% %0.1fK/s         \r",
					filename, (oldfilesize+ bytes)/1024.0, locfilesize/1024.0,
					(oldfilesize+bytes)*100/locfilesize, cps/1024 );
			}
			else
			{
				printf( "%s %ld\r", filename, oldfilesize+bytes );
			}
			fflush( stdout );
		}
	}

	t2 = TimeVal::Now() - t1;

	if( Verbose > 0 )
	{
		if( t2.GetSeconds() == 0
		 && t2.GetMicros() == 0 )
		{
			cps = 1.0;
		}
		else
		{
			cps = 1000000 * t2.GetSeconds() + t2.GetMicros();
			cps = ((double)bytes) * 1000000.0 / cps;
		}

		printf( "### %s: got %ld bytes in %ld.%06ld sec. (%ld cps)\n",
			filename, bytes, t2.GetSeconds(), t2.GetMicros(), (long)cps );
	}

	if (d < c) perror( "write" );

	close( fd );

	CloseData();

	/*
	 * Antwort vom FTP-Server: "226 Transfer complete."
	 */

	if( GetReply( 0 ) != 2 )
	{
		if (Warnings >= 0)
			fprintf( stderr, "Error retrieving file `%s': %s\n",
				filename,  ReplyBuf);
		return -1;
	}

	return 0;
}

static char *gMonName[12] =
{
	"Jan", "Feb", "Mar", "Apr", "May", "Jun",
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};

time_t tFTP::Str2Date( char *aMon, char *aDay, char *aTime )
{
	time_t tt;
	struct tm *GMT;
	int i;

	time( &tt );
	GMT = gmtime( &tt );

	GMT->tm_hour = 0;
	GMT->tm_min = 0;
	GMT->tm_sec = 0;
	GMT->tm_isdst = 0;

	for( i=0; i<12; i++ )
	{
		if( strcasecmp( aMon, gMonName[i] ) == 0 ) break;
	}
	if( i>=12 ) return 0L;

	if (i > GMT->tm_mon) (GMT->tm_year)--;

	GMT->tm_mon = i;
	GMT->tm_mday = atol( aDay );

	i = atol( aTime );
	if( i > 1900 )
	{
		GMT->tm_year = i - 1900;
	}
	else
	{
		GMT->tm_hour = 10 * (aTime[0]-'0') + (aTime[1]-'0');
		GMT->tm_min  = 10 * (aTime[3]-'0') + (aTime[4]-'0');
	}

	tt = mktime( GMT );

	return tt;
}

long tFTP::Str2Mode( char *aMode )
{
	long val = 0;
	int pos;

	switch( strlen( aMode ) )
	{
	case 10:
		for( pos=8; pos>=0; pos-- )
		{
			if( aMode[9-pos] != '-' ) val |= (1<<pos);
		}

		val &= 0777;

		switch( aMode[0] )
		{
		case 'f': val |= S_IFIFO; break;
		case 'c': val |= S_IFCHR; break;
		case 'd': val |= S_IFDIR; break;
		case 'b': val |= S_IFBLK; break;
		case '-': val |= S_IFREG; break;
		case 'l': val |= S_IFLNK; break;
		case 's': val |= S_IFSOCK; break;
		}
		break;
	}
	return val;
}

int tFTP::LocalDir( char *dirname, tCWList<tFtpDirNode> &aList )
{
	DIR *dir;
	d_type *d;
	char locfile[256];
	struct stat sbuf;
	tFtpDirNode *Node;

	if( ( dir = opendir( dirname ) ) == NULL ) return -1;

	while( ( d = readdir( dir ) ) != NULL )
	{
		if( d->d_name[0] != '.' )
		{
			sprintf( locfile, "%s/%s", dirname, d->d_name );

			if( stat( locfile, &sbuf ) == 0 )
			{
				Node = new tFtpDirNode(),

				Node->FDN_Name = mydup( d->d_name );
				Node->FDN_Size = sbuf.st_size;
				Node->FDN_Date = sbuf.st_mtime;
				Node->FDN_Mode = sbuf.st_mode;

				aList.Append( Node );
			}
		}
	}

	closedir( dir );
	return 0;
}

int tFTP::RemoteDir( char *dirname, tCWList<tFtpDirNode> &aList )
{
	int c;
	char *ptr, *last;
	char s[256];
	int ac, line = 0;
	char *av[20];
	tFtpDirNode *Node;

	DirLen = 0;
	DirPos = 0;

	if (Verbose > 0) printf( "Get Directory '%s' ...\n", dirname );

	if( Command( "TYPE A" ) != COMPLETE )
	{
		LostPeer();
		if (Warnings >= 0)
			fprintf (stderr, "Error activating ASCII transfer mode: %s\n", ReplyBuf);
		return -1;
	}

	if( InitConnection() )
	{
		Code = -1;
		if (Warnings >= 0)
			fprintf (stderr, "Error opening data connection: %s\n", ReplyBuf);
		return -1;
	}

	if( CommandData( "LIST %s", dirname ) != PRELIM )
	{
		if (Warnings >= 0)
			fprintf (stderr, "Error listing %s: %s\n", dirname, ReplyBuf);
		return -1;
	}

	/*
	 * Alloc directory buffer
	 */

	if( DirSiz == 0 )
	{
		DirSiz = 4*BUFSIZ;
		if( ( DirBuf = (char *)malloc(DirSiz) ) == NULL )
		{
			perror( "malloc" );
			Code = -1;
			CloseData();
			return -1;
		}
	}

	/*
	 * read the whole directory into the buffer first
	 */

	do
	{
		/*
		 * extend the buffer if neccessary
		 */

		if( DirLen + BUFSIZ > DirSiz )
		{
			DirSiz *= 2;
			if( ( DirBuf = (char *)realloc( DirBuf, DirSiz ) ) == NULL )
			{
				perror( "realloc" );
				Code = -1;
				CloseData();
				return -1;
			}
		}

		if( ( c = read( DataSD, &DirBuf[DirLen], BUFSIZ ) ) > 0 )
		{
			DirLen += c;
		}
	} while( c > 0 );

	CloseData();

	/*
	 * Antwort: "226 Transfer complete."
	 */

	if( GetReply( 0 ) != 2 )
	{
		if (Warnings >= 0)
			fprintf( stderr, "Error retrieving directory `%s': %s\n",
				dirname,  ReplyBuf);
		return -1;
	}

	DirBuf[DirLen]='\0';
	ptr = DirBuf;
	last = DirBuf;

	while( *ptr != '\0' )
	{
		if( *ptr == '\n' || *ptr == '\r' )
		{
			*ptr++='\0';

			if( strlen( last ) == 0 )
			{
				last = ptr;
				continue;
			}

			line++;

			if( Verbose > 2 )
			{
				printf( "dir %s\n", last );
			}

			strcpy (s, last);
			ac = SplitString( last, av, NULL );

			if( line == 1 && LsType == LSTYPE_UNKNOWN)
			{
				if( ac>0 && strcmp( av[0], "total" ) == 0 )
				{
					LsType = LSTYPE_UNIX;
				}
				else if( ac>0 && strcmp( av[0], "directory" ) == 0 )
				{
					LsType = LSTYPE_HP;
				}
				else
				{
					if (Warnings >= 1)
						fprintf( stderr, "Warning: system type unknown, assuming UN*X.\n" );
					LsType = LSTYPE_UNIX;
				}
			}
			else
			{
				Node = new tFtpDirNode();
				aList.Append( Node );

				switch( LsType )
				{
				case LSTYPE_UNIX:
					if( ac == 9 )
					{
						/* 0=mode, 1=links, 2=owner, 3=group */
						/* 4=size, 5=Month, 6=Day */
						/* 7=Time/Year, 8=Name */

						Node->FDN_Name = mydup( av[8] );
						Node->FDN_Size = atol( av[4] );
						Node->FDN_Date = Str2Date( av[5], av[6], av[7] );
						Node->FDN_Mode = Str2Mode( av[0] );
					}
					else if( ac == 8 )
					{
						/* 0=mode, 1=links, 2=owner */
						/* 3=size, 4=Month, 5=Day */
						/* 6=Time/Year, 7=Name */

						Node->FDN_Name = mydup( av[7] );
						Node->FDN_Size = atol( av[3] );
						Node->FDN_Date = Str2Date( av[4], av[5], av[6] );
						Node->FDN_Mode = Str2Mode( av[0] );
					}
					else if( ac != 2) // "total 123"
					{
						if (Warnings >= 1)
							fprintf( stderr, "Warning: can't read dir entry `%s', ignored\n",
								s );
					}
					break;
				case LSTYPE_HP:
					if( ac > 5
					 && ( strcmp( av[1], "rw/rw/rw" ) == 0 
					   || strcmp( av[2], "rw/rw/rw" ) == 0 ) )
					{
						Node->FDN_Name = mydup( av[0] );
						Node->FDN_Mode = S_IFREG;
					}
					else
					{
						if (Warnings >= 1)
							fprintf( stderr, "Warning: can't read dir entry `%s', ignored\n",
								s );
					}
					break;
				case LSTYPE_MACOS:
					if( ac == 8 )
					{
						/* 0=mode, 1=owner, 2=group */
						/* 3=size, 4=Month, 5=Day */
						/* 6=Time/Year, 7=Name */

						Node->FDN_Name = mydup( av[7] );
						Node->FDN_Size = atol( av[3] );
						Node->FDN_Date = Str2Date( av[4], av[5], av[6] );
						Node->FDN_Mode = Str2Mode( av[0] );
					}
					else if( ac == 7 )
					{
						/* 0=mode, 1='folder' */
						/* 2=size, 3=Month, 4=Day */
						/* 5=Time/Year, 6=Name */

						Node->FDN_Name = mydup( av[6] );
						Node->FDN_Size = atol( av[2] );
						Node->FDN_Date = Str2Date( av[3], av[4], av[5] );
						Node->FDN_Mode = Str2Mode( av[0] );
					}
					else
					{
						if (Warnings >= 1)
							fprintf( stderr, "Warning: can't read dir entry `%s', ignored\n",
								s );
					}
					break;
				default:
					if (Warnings >= 1)
						fprintf (stderr, "Error: Can't determine system type, directory not read.\n");
					return -1;
				}
			}

			last = ptr;
		}
		else
		{
			ptr++;
		}
	}

	return 0;
}

long tFTP::FileSize( char *aFilename )
{
	char *Ptr;

	if( Command( "SIZE %s", aFilename ) != COMPLETE )
	{
		return -1;
	}
	if( ( Ptr = strchr( ReplyBuf, ' ' ) ) != NULL )
	{
		Ptr++;
		return atol( Ptr );
	}
	return -1;
}

#define VAL( b, i ) ( (b[i]-'0')*10+b[i+1]-'0' )

time_t tFTP::FileTime( char *aFilename )
{
	char *Ptr;
	struct tm *lt;
	time_t tt;

	if( Command( "MDTM %s", aFilename ) != COMPLETE )
	{
		return -1;
	}

	if( ( Ptr = strchr( ReplyBuf, ' ' ) ) != NULL )
	{
		/*        01234567890123 */
		Ptr++; /* 19951119230048 */

		time(&tt);
		lt = localtime( &tt );
		lt->tm_year = VAL( Ptr, 2 );
		lt->tm_mon  = VAL( Ptr, 4 ) - 1;
		lt->tm_mday = VAL( Ptr, 6 );
		lt->tm_hour = VAL( Ptr, 8 );
		lt->tm_min  = VAL( Ptr, 10 );
		lt->tm_sec  = VAL( Ptr, 12 );

		return mktime( lt );
	}
	return -1;
}
