/*
 *  Ftp.C - Generic FTP C++ class (definition)
 *  Copyright (C) 1995 Christian Wolf
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *  You can reach me, the author of cwftp as <chris@MedIS.DE>
 *
 * $Id: Ftp.H,v 1.3 1996/05/13 16:49:54 chris Exp $
 *
 * $Log: Ftp.H,v $
 * Revision 1.3  1996/05/13 16:49:54  chris
 * STOR, Bugfixes
 *
 * Revision 1.2  1996/01/30  09:26:04  chris
 * *** empty log message ***
 *
 * Revision 1.1  1995/05/08  11:47:52  chris
 * Initial revision
 *
 *
 */

#ifndef FTP_H
#define FTP_H

#include <sys/types.h>
#include <time.h>
#include <netinet/in.h>
#include <stdio.h>

#include "CWList.H"

struct tFtpDirNode
{
	char * FDN_Name; /* filename */
	long   FDN_Size; /* filesize, -1 when unknown */
	time_t FDN_Date; /* timestamp, 0 when unknown */
	int    FDN_Mode; /* filemode,  0 when unknown */

	tFtpDirNode();
	~tFtpDirNode();
	char *MakeStr();
};

class tFTP
{
	private:
		int CommandSD;		/* Command Socket-Descriptor */
		int BindSD;		/* Bind Socket-Descriptor */
		int DataSD;		/* Data Socket-Descriptor */
		int Connected;
		int Code;
		struct sockaddr_in MyClientAddress;
		struct sockaddr_in DataAddress;
		char InBuf[BUFSIZ];
		char ReplyBuf[BUFSIZ];
		int InLen;
		int InPos;
		char *DirBuf;
		int DirLen;
		int DirSiz;
		int DirPos;
		int Verbose;
		int Warnings;
		int UsePASV;
		int LsType;

	protected:
		time_t Str2Date( char *aMon, char *aDay, char *aTime );
		long Str2Mode( char *aMode );
		char *HErrorStr( int err );

		int GetC();
		void LostPeer();
		void CloseData();

		int InitConnection();
		int GetReply( int waiteof );
		void SendCommand( char *fmt, ... );
		int Command( char *fmt, ... );
		int CommandData( char *fmt, ... );
		int GetSystemType ();
		int DataConnection();

	public:
		tFTP( int aVerbose=0, int aWarnings=0 );
		~tFTP();

		int IsConnected() { return Connected; };

		int Open( char *aHostname, char *aUsername="ftp", char *aPassword=NULL );
		int Close();

		int ChangeDir( char *aDirname );
		int Get( char *aFilename, char *aLocalname, long aFileSize, int asAscii=0 );
		int Put( char *aFilename, char *aLocalname, int asAscii=0 );
		int RemoteDir( char *aDirname, tCWList<tFtpDirNode> &aList );
		int LocalDir( char *aDirname, tCWList<tFtpDirNode> &aList );
		void FreeDir( tCWList<tFtpDirNode> &aList );
		long FileSize( char *aFilename );
		time_t FileTime( char *aFilename );
};

#endif

