/*
 *  RegExp.C - Generic regular expression C++ class (implementation)
 *  Copyright (C) 1995 Christian Wolf
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *  You can reach me, the author of cwftp as <chris@MedIS.DE>
 *
 * $Id: RegExp.C,v 1.3 1996/03/10 13:41:23 chris Exp $
 *
 * $Log: RegExp.C,v $
 * Revision 1.3  1996/03/10  13:41:23  chris
 * *** empty log message ***
 *
 * Revision 1.2  1996/01/30  09:26:04  chris
 * *** empty log message ***
 *
 * Revision 1.1  1995/05/08  11:47:52  chris
 * Initial revision
 *
 *
 */

#include "RegExp.H"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef HAVE_REGEX

/*
 * Der INIT Makro steht am Anfang der compile() Fukntion. "instring"
 * ist der erste Parameter dieser Funktion.
 */

#define INIT         register char *sp = instring;

/*
 * Der GETC() Makro wird fuer jedes Zeichen aufgerufen
 */

#define GETC()       (*sp++)

/*
 * Der PEEKC() Makro sieht sich das naechste Zeichen an, ohne es aus
 * dem "Keller" zu nehmen
 */

#define PEEKC()      (*sp)

/*
 * Der UNGETC() Makro schreibt ei Zeichen wieder in den Keller zurueck
 * und zwar immer nur das Zeichen, welches schon vorher dort stand
 */

#define UNGETC(ch)   (--sp)

/*
 * Dieser Makro wird bei einem match aufgerufen
 */

#define RETURN(ptr)  return( NULL )

/*
 * Dieser Makro wird bei einem Fehler in der Expression aufgerufen
 */

#define ERROR(val)   return( (char*)val )

/*
 * Hier werden die Funktionen compile(), step() und advance()
 * definiert.
 */

#include <regexp.h>

#endif

tRegularExpression :: tRegularExpression( char *aRegExpStr )
{
	ExpPattern = NULL;

	SetNewExpression( aRegExpStr );
}

tRegularExpression :: ~tRegularExpression()
{
	SetNewExpression( NULL );
}

int tRegularExpression :: SetNewExpression( char *aRegExpStr )
{
	if( ExpPattern != NULL )
	{
		free( ExpPattern );
#ifdef HAVE_REGEX
		regfree( &ExpPre );
#endif
	}

	ExpError = -1;
	ExpPattern = NULL;

	if( aRegExpStr != NULL )
	{
		ExpPattern = strdup( aRegExpStr );
#ifdef HAVE_REGEX
		ExpError = regcomp( &ExpPre, ExpPattern, REG_NOSUB );
#else
		ExpError = (int) compile( aRegExpStr, ExpBuffer, &ExpBuffer[MAX_REG], '\0' );
#endif
	}
	return ExpError;
}

int tRegularExpression :: Match( char *aTextStr )
{
	if( !ExpPattern ) return 0;
	if( ExpError == 0 )
	{
#ifdef HAVE_REGEX
		return regexec( &ExpPre, aTextStr, 0, NULL, 0 );
#else
		return step( aTextStr, ExpBuffer );
#endif
	}
	else
	{
		return 0;
	}
}

#ifdef SELF_MAIN
int main( int argc, char **argv )
{
	char tstbuf[ BUFSIZ ];

	tRegularExpression aRE( argv[1] );

	while( gets( tstbuf ) != NULL )
	{
		if( aRE.Match( tstbuf ) )
		{
			printf( "TRUE\n" );
		}
	}

	return 0;
}
#endif
