/*
 *  TimeVal.H - TimeVal interface class (definition, no implementation)
 *  Copyright (C) 1995 Christian Wolf
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *  You can reach me, the author of cwftp as <chris@MedIS.DE>
 *
 * $Id: TimeVal.H,v 1.1 1995/05/08 11:47:52 chris Exp $
 *
 * $Log: TimeVal.H,v $
 * Revision 1.1  1995/05/08  11:47:52  chris
 * Initial revision
 *
 * Revision 1.1  1995/05/08  11:47:52  chris
 * Initial revision
 *
 *
 */

#ifndef TimeVal_H
#define TimeVal_H

#include <time.h>
#include <sys/time.h>

class TimeVal : public timeval
{
	// friend class TimeVal;

public:
	/* Konstrkutoren */

	TimeVal() { tv_sec = 0; tv_usec = 0; }
	TimeVal( time_t aTime ) { tv_sec = aTime; tv_usec = 0; }
	TimeVal( long aSec, long aUSec ) { tv_sec = aSec; tv_usec = aUSec; }

	/* Werte */

	long GetSeconds() { return tv_sec; }
	long GetMicros() { return tv_usec; }
	time_t GetTime() { return tv_sec; }

	/* Methoden */

	void SetNow() { gettimeofday( this, NULL ); }

	/* Current-Time */

	static TimeVal Now() { TimeVal tv; gettimeofday( &tv, NULL ); return tv; }

	/* Operatoren */

	int operator== ( TimeVal &aOtherVal )
	{
		return (
			tv_sec == aOtherVal.tv_sec &&
			tv_usec == aOtherVal.tv_usec );
	}

	int operator!= ( TimeVal &aOtherVal )
	{
		return (
			tv_sec != aOtherVal.tv_sec ||
			tv_usec != aOtherVal.tv_usec );
	}

	int operator<( TimeVal &aOtherVal )
	{
		if( tv_sec < aOtherVal.tv_sec ||
		    ( tv_sec == aOtherVal.tv_sec &&
		      tv_usec < aOtherVal.tv_usec ) )
		{
			return 1;
		}
		return 0;
	}

	int operator>( TimeVal &aOtherVal )
	{
		if( tv_sec > aOtherVal.tv_sec ||
		    ( tv_sec == aOtherVal.tv_sec &&
		      tv_usec > aOtherVal.tv_usec ) )
		{
			return 1;
		}
		return 0;
	}

	int operator>=( TimeVal &aOtherVal )
	{
		if( tv_sec >= aOtherVal.tv_sec ||
		    ( tv_sec == aOtherVal.tv_sec &&
		      tv_usec >= aOtherVal.tv_usec ) )
		{
			return 1;
		}
		return 0;
	}

	int operator<=( TimeVal &aOtherVal )
	{
		if( tv_sec <= aOtherVal.tv_sec ||
		    ( tv_sec == aOtherVal.tv_sec &&
		      tv_usec <= aOtherVal.tv_usec ) )
		{
			return 1;
		}
		return 0;
	}

	TimeVal operator+( TimeVal &aOtherVal )
	{
		TimeVal tv;

		tv.tv_sec = tv_sec + aOtherVal.tv_sec;
		tv.tv_usec = tv_usec + aOtherVal.tv_usec;

		if( tv.tv_usec > 1000000 )
		{
			tv.tv_usec -= 1000000;
			tv.tv_sec += 1;
		}
		return tv;
	}

	TimeVal operator-( TimeVal &aOtherVal )
	{
		TimeVal tv;

		tv.tv_sec = tv_sec - aOtherVal.tv_sec;
		tv.tv_usec = tv_usec - aOtherVal.tv_usec;

		if( tv.tv_usec < 0 )
		{
			tv.tv_usec += 1000000;
			tv.tv_sec -= 1;
		}
		return tv;
	}
};

#endif /* TimeVal_H */
